async def create_table(dbsession):
    async with dbsession.cursor() as cursor:
        stmt = """
        CREATE TABLE IF NOT EXISTS event_store (
            id varchar(255),
            data jsonb NOT NULL,
            created_at timestamp NOT NULL,
            updated_at timestamp NOT NULL
        )
        """
        await cursor.execute(stmt)


async def drop_table(dbsession):
    async with dbsession.cursor() as cursor:
        stmt = """
        DROP TABLE event_store
        """
        await cursor.execute(stmt)



from kant.events.models import EventModel


class Query:
    def __init__(self):
        self._where = []
        self._parameters = {}

    def __call__(self):
        operation = 'SELECT * FROM event_store'
        operation += self._stmt_parameters()
        return (operation, self._parameters)

    def _stmt_parameters(self):
        stmt = ''
        stmt_parameters = []
        for (name, value, lookup,) in self._where:
            if isinstance(value, EventModel):
                stmt_param = self._clean_event(name, value, lookup)
            elif isinstance(value, list):
                stmt_param = self._clean_list(name, value, lookup)
            else:
                stmt_param = self._clean_default(name, value, lookup)
            stmt_parameters.append(stmt_param)

        if len(stmt_parameters) > 0:
            stmt = ' WHERE ' + ' AND '.join(stmt_parameters)
        return stmt

    def _clean_event(self, name, value, lookup):
        self._parameters[name] = value.decode()
        stmt_param = '{0} {1} %({0})s::jsonb'.format(name, lookup)
        return stmt_param

    def _clean_list(self, name, value, lookup):
        self._parameters[name] = value
        stmt_param = '{0} {1} (%({0})s)'.format(name, lookup)
        return stmt_param

    def _clean_default(self, name, value, lookup):
        self._parameters[name] = value
        stmt_param = '{0} {1} %({0})s'.format(name, lookup)
        return stmt_param

    def filter(self, *args, **kwargs):
        for name, value in kwargs.items():
            operator = '='
            if isinstance(value, EventModel):
                operator = '@>'
            elif isinstance(value, list):
                operator = 'IN'
            self._where.append((name, value, operator))
        return self



from operator import attrgetter
from datetime import datetime
import json

from kant.exceptions import VersionError, ObjectDoesNotExist, StreamExists
from kant.events.serializers import EventModelEncoder
from kant.events import EventModel
from kant.eventstore.stream import EventStream


class EventStoreRepository:
    def __init__(self, session, *args, **kwargs):
        self.session = session

    async def update(self, aggregate_id, events: EventStream):
        async with self.session.begin() as transaction:
            stmt_select = """
            SELECT event_store.data
            FROM event_store WHERE event_store.id = %(id)s AND CAST(data ? '$version' AS INTEGER) >= %(version)s
            FOR UPDATE
            """
            await self.session.execute(stmt_select, {
                'id': str(aggregate_id),
                'version': events.initial_version,
            })
            event_store = await self.session.fetchone()
            if not event_store:
                raise ObjectDoesNotExist()
            stored_events = EventStream.make(event_store[0])
            if stored_events > events:
                message = "The version is {current_version}, but the expected is {expected_version}".format(
                    current_version=stored_events.initial_version,
                    expected_version=events.initial_version,
                )
                raise VersionError(message)
            stored_events += events
            stmt_update = """
            UPDATE event_store SET data=%(data)s, updated_at=NOW() WHERE id = %(id)s;
            """
            await self.session.execute(stmt_update, {
                'id': str(aggregate_id),
                'data': stored_events.decode(),
            })

    async def create(self, aggregate_id, events: EventStream):
        stmt = """
        INSERT INTO event_store (id, data, created_at, updated_at)
        VALUES (%(id)s, %(data)s, NOW(), NOW())
        """
        await self.session.execute(stmt, {
            'id': str(aggregate_id),
            'data': events.decode(),
        })

    async def save(self, aggregate_id, events: EventStream):
        """ Save the events in the model """
        try:
            await self.update(aggregate_id, events)
        except ObjectDoesNotExist:
            await self.create(aggregate_id, events)
        return events.current_version

    async def get(self, aggregate_id, initial_version=0):
        stmt = """
        SELECT event_store.data
        FROM event_store WHERE event_store.id = %(id)s AND CAST(data ? '$version' AS INTEGER) >= %(version)s
        """
        await self.session.execute(stmt, {
            'id': str(aggregate_id),
            'version': initial_version,
        })
        event_store = await self.session.fetchone()
        if not event_store:
            raise ObjectDoesNotExist()
        return EventStream.make(event_store[0])



from json import JSONEncoder, JSONDecoder
from .models import EventModel


class EventModelEncoder(JSONEncoder):
    """
    A class serializer for EventModel to be converted to json
    >>> found_added = FoundAdded(amount=25.5)
    >>> isinstance(found_added, EventModel)
    True
    >>> json.dumps(found_added, cls=EventModelEncoder, sort_keys=True)
    '{"$type": "FoundAdded", "$version": 0, "amount": 25.5}'
    """
    def default(self, obj):
        if isinstance(obj, EventModel):
            return obj.decode()
        return JSONEncoder.default(self, obj)



from copy import deepcopy
import json
from inflection import underscore
from kant.core.datamapper.base import ModelMeta, FieldMapping
from kant.eventstore.stream import EventStream


class Aggregate(FieldMapping, metaclass=ModelMeta):
    def __init__(self):
        super().__init__()
        self._all_events = EventStream()
        self._events = EventStream()
        self._stored_events = EventStream()

    def all_events(self):
        return self._all_events

    def stored_events(self):
        return self._stored_events

    def get_events(self):
        return self._events

    def clear_events(self):
        self._events = EventStream()

    def fetch_events(self, events: EventStream):
        self._stored_events = deepcopy(events)
        self._all_events = deepcopy(self._stored_events)
        for event in events:
            self.dispatch(event, flush=False)

    def apply(self, event):
        event_name = underscore(event.__class__.__name__)
        method_name = 'apply_{0}'.format(event_name)
        method = getattr(self, method_name)
        method(event)

    def dispatch(self, event, flush=True):
        self.apply(event)
        if flush:
            self._events.add(event)
            self._all_events.add(event)

    def json(self):
        data = {key: value for key, value in self.serializeditems()}
        return json.dumps(data, sort_keys=True)

    @property
    def current_version(self):
        return self._all_events.current_version

    @property
    def version(self):
        return self._all_events.initial_version



from datetime import datetime
import json
import pytest
from tests.fixtures import AccountSchemaModel, FoundAdded


@pytest.fixture(autouse=True)
def event_model_example(doctest_namespace):
    doctest_namespace['FoundAdded'] = FoundAdded
    doctest_namespace['json'] = json
    doctest_namespace['datetime'] = datetime
    doctest_namespace['AccountSchemaModel'] = AccountSchemaModel
